@@ -21,7 +21,6 @@
 #if ENABLE(BUBBLEWRAP_SANDBOX)
 
 #include "XDGDBusProxy.h"
-#include <WebCore/PlatformDisplay.h>
 #include <fcntl.h>
 #include <glib.h>
 #include <seccomp.h>
@@ -31,8 +30,15 @@
 #include <wtf/NeverDestroyed.h>
 #include <wtf/UUID.h>
 #include <wtf/UniStdExtras.h>
+#include <wtf/glib/Application.h>
 #include <wtf/glib/GRefPtr.h>
 #include <wtf/glib/GUniquePtr.h>
+#include <wtf/glib/Sandbox.h>
+#include <wtf/text/MakeString.h>
+
+#if PLATFORM(GTK)
+#include "Display.h"
+#endif
 
 #if !defined(MFD_ALLOW_SEALING) && HAVE(LINUX_MEMFD_H)
 #include <linux/memfd.h>
@@ -125,33 +131,6 @@ int argumentsToFileDescriptor(const Vector<CString>& args, const char* name)
     return memfd;
 }
 
-static String effectiveApplicationId()
-{
-    if (auto* app = g_application_get_default()) {
-        if (const char* appID = g_application_get_application_id(app))
-            return String::fromUTF8(appID);
-    }
-
-    const char* programName = g_get_prgname();
-    if (programName && g_application_id_is_valid(programName))
-        return String::fromUTF8(programName);
-
-    // There must be some id for xdg-desktop-portal to function.
-    // xdg-desktop-portal uses this id for permissions.
-    // This creates a somewhat reliable id based on the executable path
-    // which will avoid potentially gaining permissions from another app
-    // and won't flood xdg-desktop-portal with new ids.
-    if (auto executablePath = FileSystem::currentExecutablePath(); !executablePath.isNull()) {
-        GUniquePtr<char> digest(g_compute_checksum_for_data(G_CHECKSUM_SHA256, reinterpret_cast<const uint8_t*>(executablePath.data()), executablePath.length()));
-        return makeString("org.webkit.app-", digest.get());
-    }
-
-    // If it is not possible to obtain the executable path, generate
-    // a random identifier as a fallback.
-    auto uuid = WTF::UUID::createVersion4Weak();
-    return makeString("org.webkit.app-", uuid.toString());
-}
-
 static void createBwrapInfo(GSubprocessLauncher* launcher, Vector<CString>& args, const char* instanceID)
 {
     // This is the hardcoded path expected in xdg-desktop-portal's xdp_app_info_load_bwrap_info() used
@@ -182,7 +161,7 @@ static int createFlatpakInfo(const char* instanceID)
 
     if (!data.get()) {
         GUniquePtr<GKeyFile> keyFile(g_key_file_new());
-        g_key_file_set_string(keyFile.get(), "Application", "name", effectiveApplicationId().utf8().data());
+        g_key_file_set_string(keyFile.get(), "Application", "name", WTF::applicationID().data());
         g_key_file_set_string(keyFile.get(), "Instance", "instance-id", instanceID);
         data->reset(g_key_file_to_data(keyFile.get(), &size, nullptr));
     }
@@ -229,19 +208,13 @@ static void bindIfExists(Vector<CString>& args, const char* path, BindFlags bind
         args.appendVector(Vector<CString>({ bindType, path, path }));
 }
 
-static const char* dbusProxyDirectory()
-{
-    static GUniquePtr<char> path(g_build_filename(g_get_user_runtime_dir(), BASE_DIRECTORY, nullptr));
-    return path.get();
-}
-
 static void bindDBusSession(Vector<CString>& args, XDGDBusProxy& dbusProxy, bool allowPortals)
 {
     auto dbusSessionProxyPath = dbusProxy.dbusSessionProxy(BASE_DIRECTORY, allowPortals ? XDGDBusProxy::AllowPortals::Yes : XDGDBusProxy::AllowPortals::No);
     if (!dbusSessionProxyPath)
         return;
 
-    GUniquePtr<char> sandboxedSessionBusPath(g_build_filename(dbusProxyDirectory(), "bus", nullptr));
+    GUniquePtr<char> sandboxedSessionBusPath(g_build_filename(sandboxedUserRuntimeDirectory().data(), "bus", nullptr));
     GUniquePtr<char> proxyAddress(g_strdup_printf("unix:path=%s", sandboxedSessionBusPath.get()));
     args.appendVector(Vector<CString> {
         "--ro-bind", *dbusSessionProxyPath, sandboxedSessionBusPath.get(),
@@ -380,19 +353,21 @@ static void bindGtkData(Vector<CString>& args)
 }
 #endif
 
-static void bindA11y(Vector<CString>& args, XDGDBusProxy& dbusProxy)
+#if USE(ATSPI)
+static void bindA11y(Vector<CString>& args, XDGDBusProxy& dbusProxy, const String& accessibilityBusAddress, const String& accessibilityBusName, const String& sandboxedAccessibilityBusAddress)
 {
-    GUniquePtr<char> sandboxedAccessibilityBusPath(g_build_filename(dbusProxyDirectory(), "at-spi-bus", nullptr));
-    auto accessibilityProxyPath = dbusProxy.accessibilityProxy(BASE_DIRECTORY, sandboxedAccessibilityBusPath.get());
+    auto accessibilityProxyPath = dbusProxy.accessibilityProxy(BASE_DIRECTORY, accessibilityBusAddress, accessibilityBusName);
     if (!accessibilityProxyPath)
         return;
 
-    GUniquePtr<char> proxyAddress(g_strdup_printf("unix:path=%s", sandboxedAccessibilityBusPath.get()));
+    ASSERT(sandboxedAccessibilityBusAddress.startsWith("unix:path="_s));
+    auto sandboxedAccessibilityBusPath = sandboxedAccessibilityBusAddress.substring(strlen("unix:path="));
     args.appendVector(Vector<CString> {
-        "--ro-bind", *accessibilityProxyPath, sandboxedAccessibilityBusPath.get(),
-        "--setenv", "AT_SPI_BUS_ADDRESS", proxyAddress.get()
+        "--ro-bind", *accessibilityProxyPath, sandboxedAccessibilityBusPath.utf8(),
+        "--setenv", "AT_SPI_BUS_ADDRESS", sandboxedAccessibilityBusAddress.utf8(),
     });
 }
+#endif
 
 static bool bindPathVar(Vector<CString>& args, const char* varname)
 {
@@ -702,11 +677,11 @@ static bool shouldUnshareNetwork(ProcessLauncher::ProcessType processType, char*
     if (processType == ProcessLauncher::ProcessType::DBusProxy)
         return false;
 
-#if PLATFORM(X11)
+#if PLATFORM(GTK)
     // Also, the web process needs access to host networking if the X server is running over TCP or
     // on a different host's Unix socket; this is likely the case if the first character of DISPLAY
     // is not a colon.
-    if (processType == ProcessLauncher::ProcessType::Web && PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::X11) {
+    if (processType == ProcessLauncher::ProcessType::Web && Display::singleton().isX11()) {
         const char* display = g_getenv("DISPLAY");
         if (display && display[0] != ':')
             return false;
@@ -730,7 +705,7 @@ static std::optional<CString> directoryContainingDBusSocket(const char* dbusAddr
         while (*pathEnd && *pathEnd != ',')
             pathEnd++;
 
-        CString path(pathStart, pathEnd - pathStart);
+        CString path({ pathStart, pathEnd });
         GRefPtr<GFile> file = adoptGRef(g_file_new_for_path(path.data()));
         GRefPtr<GFile> parent = adoptGRef(g_file_get_parent(file.get()));
         if (!parent)
@@ -752,7 +727,7 @@ static void addExtraPaths(const HashMap<CString, SandboxPermission>& paths, Vect
     }
 }
 
-GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const ProcessLauncher::LaunchOptions& launchOptions, char** argv, GError **error)
+GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const ProcessLauncher::LaunchOptions& launchOptions, XDGDBusProxy& dbusProxy, char** argv, GError **error)
 {
     ASSERT(launcher);
 
@@ -764,7 +739,6 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
 
     const char* runDir = g_get_user_runtime_dir();
     Vector<CString> sandboxArgs = {
-        "--die-with-parent",
         "--unshare-uts",
 
         // We assume /etc has safe permissions.
@@ -830,7 +804,7 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
     if (launchOptions.processType == ProcessLauncher::ProcessType::DBusProxy) {
         sandboxArgs.appendVector(Vector<CString>({
             "--ro-bind", DBUS_PROXY_EXECUTABLE, DBUS_PROXY_EXECUTABLE,
-            "--bind", dbusProxyDirectory(), dbusProxyDirectory(),
+            "--bind", sandboxedUserRuntimeDirectory().data(), sandboxedUserRuntimeDirectory().data(),
         }));
 
         // xdg-dbus-proxy is trusted, so it's OK to mount the directories that contain the session
@@ -843,11 +817,13 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
             }));
         }
 
-        if (auto a11yBusDirectory = directoryContainingDBusSocket(PlatformDisplay::sharedDisplay().accessibilityBusAddress().utf8().data())) {
+#if USE(ATSPI)
+        if (auto a11yBusDirectory = directoryContainingDBusSocket(launchOptions.extraInitializationData.get("accessibilityBusAddress"_s).utf8().data())) {
             sandboxArgs.appendVector(Vector<CString>({
                 "--bind", *a11yBusDirectory, *a11yBusDirectory,
             }));
         }
+#endif
     }
 
     if (shouldUnshareNetwork(launchOptions.processType, argv))
@@ -883,30 +859,33 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
         }));
     }
 
+    bindIfExists(sandboxArgs, "/run/systemd/journal/socket");
+    bindIfExists(sandboxArgs, "/run/systemd/journal/stdout");
+
     createBwrapInfo(launcher, sandboxArgs, instanceID.get());
 
     if (launchOptions.processType == ProcessLauncher::ProcessType::Web) {
+#if PLATFORM(GTK)
 #if PLATFORM(WAYLAND)
-        if (PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::Wayland) {
+        if (Display::singleton().isWayland()) {
             bindWayland(sandboxArgs);
             sandboxArgs.append("--unshare-ipc");
         }
 #endif
 #if PLATFORM(X11)
-        if (PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::X11)
+        if (Display::singleton().isX11())
             bindX11(sandboxArgs);
+#endif
 #endif
 
-        Vector<String> extraPaths = { "applicationCacheDirectory"_s, "mediaKeysDirectory"_s, "waylandSocket"_s };
+        Vector<String> extraPaths = { "mediaKeysDirectory"_s, "waylandSocket"_s };
         for (const auto& path : extraPaths) {
             String extraPath = launchOptions.extraInitializationData.get(path);
             if (!extraPath.isEmpty())
                 sandboxArgs.appendVector(Vector<CString>({ "--bind-try", extraPath.utf8(), extraPath.utf8() }));
         }
 
-        static std::unique_ptr<XDGDBusProxy> dbusProxy = makeUnique<XDGDBusProxy>();
-        if (dbusProxy)
-            bindDBusSession(sandboxArgs, *dbusProxy, flatpakInfoFd != -1);
+        bindDBusSession(sandboxArgs, dbusProxy, flatpakInfoFd != -1);
         // FIXME: We should move to Pipewire as soon as viable, Pulse doesn't restrict clients atm.
         bindPulse(sandboxArgs);
         bindSndio(sandboxArgs);
@@ -915,13 +894,18 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
         bindOpenGL(sandboxArgs);
         // FIXME: This is also fixed by Pipewire once in use.
         bindV4l(sandboxArgs);
-        if (dbusProxy)
-            bindA11y(sandboxArgs, *dbusProxy);
+#if USE(ATSPI)
+        auto accessibilityBusAddress = launchOptions.extraInitializationData.get("accessibilityBusAddress"_s);
+        auto sandboxedAccessibilityBusAddress = launchOptions.extraInitializationData.get("sandboxedAccessibilityBusAddress"_s);
+        if (!accessibilityBusAddress.isEmpty() && !sandboxedAccessibilityBusAddress.isEmpty()) {
+            auto a11yBusName = launchOptions.extraInitializationData.get<HashTranslatorASCIILiteral>("accessibilityBusName"_s);
+            bindA11y(sandboxArgs, dbusProxy, accessibilityBusAddress, a11yBusName, sandboxedAccessibilityBusAddress);
+        }
+#endif
 #if PLATFORM(GTK)
         bindGtkData(sandboxArgs);
 #endif
-        if (dbusProxy && !dbusProxy->launch())
-            dbusProxy = nullptr;
+        dbusProxy.launch(launchOptions);
     } else {
         // Only X11 users need this for XShm which is only the Web process.
         sandboxArgs.append("--unshare-ipc");
