@@ -26,10 +26,10 @@
 #include "config.h"
 #include "WPEDisplay.h"
 
-#include "WPEBufferDMABufFormat.h"
 #include "WPEDisplayPrivate.h"
 #include "WPEEGLError.h"
 #include "WPEExtensions.h"
+#include "WPEInputMethodContextNone.h"
 #include <epoxy/egl.h>
 #include <gio/gio.h>
 #include <mutex>
@@ -37,6 +37,7 @@
 #include <wtf/NeverDestroyed.h>
 #include <wtf/glib/GRefPtr.h>
 #include <wtf/glib/GUniquePtr.h>
+#include <wtf/glib/GWeakPtr.h>
 #include <wtf/glib/WTFGType.h>
 #include <wtf/text/CString.h>
 #include <wtf/text/StringHash.h>
@@ -56,14 +57,15 @@ struct _WPEDisplayPrivate {
     EGLDisplay eglDisplay;
     GUniqueOutPtr<GError> eglDisplayError;
     HashMap<String, bool> extensionsMap;
-    GList* preferredDMABufFormats;
+    GRefPtr<WPEBufferDMABufFormats> preferredDMABufFormats;
+    GRefPtr<WPEKeymap> keymap;
 };
 
 WEBKIT_DEFINE_ABSTRACT_TYPE(WPEDisplay, wpe_display, G_TYPE_OBJECT)
 
 enum {
-    MONITOR_ADDED,
-    MONITOR_REMOVED,
+    SCREEN_ADDED,
+    SCREEN_REMOVED,
 
     LAST_SIGNAL
 };
@@ -79,11 +81,20 @@ static guint signals[LAST_SIGNAL] = { 0, };
  **/
 G_DEFINE_QUARK(wpe-display-error-quark, wpe_display_error)
 
+static GWeakPtr<WPEDisplay> s_primaryDisplay;
+
+static void wpeDisplayConstructed(GObject* object)
+{
+    if (!s_primaryDisplay)
+        s_primaryDisplay.reset(WPE_DISPLAY(object));
+
+    G_OBJECT_CLASS(wpe_display_parent_class)->constructed(object);
+}
+
 static void wpeDisplayDispose(GObject* object)
 {
     auto* priv = WPE_DISPLAY(object)->priv;
 
-    g_clear_list(&priv->preferredDMABufFormats, reinterpret_cast<GDestroyNotify>(wpe_buffer_dma_buf_format_free));
     g_clear_pointer(&priv->eglDisplay, eglTerminate);
 
     G_OBJECT_CLASS(wpe_display_parent_class)->dispose(object);
@@ -92,46 +103,47 @@ static void wpeDisplayDispose(GObject* object)
 static void wpe_display_class_init(WPEDisplayClass* displayClass)
 {
     GObjectClass* objectClass = G_OBJECT_CLASS(displayClass);
+    objectClass->constructed = wpeDisplayConstructed;
     objectClass->dispose = wpeDisplayDispose;
 
     /**
-     * WPEDisplay::monitor-added:
+     * WPEDisplay::screen-added:
      * @display: a #WPEDisplay
-     * @monitor: the #WPEMonitor added
+     * @screen: the #WPEScreen added
      *
-     * Emitted when a monitor is added
+     * Emitted when a screen is added
      */
-    signals[MONITOR_ADDED] = g_signal_new(
-        "monitor-added",
+    signals[SCREEN_ADDED] = g_signal_new(
+        "screen-added",
         G_TYPE_FROM_CLASS(displayClass),
         G_SIGNAL_RUN_LAST,
         0, nullptr, nullptr,
         g_cclosure_marshal_generic,
         G_TYPE_NONE, 1,
-        WPE_TYPE_MONITOR);
+        WPE_TYPE_SCREEN);
 
     /**
-     * WPEDisplay::monitor-removed:
+     * WPEDisplay::screen-removed:
      * @display: a #WPEDisplay
-     * @monitor: the #WPEMonitor removed
+     * @screen: the #WPEScreen removed
      *
-     * Emitted after a monitor is removed.
-     * Note that the monitor is always invalidated before this signal is emitted.
+     * Emitted after a screen is removed.
+     * Note that the screen is always invalidated before this signal is emitted.
      */
-    signals[MONITOR_REMOVED] = g_signal_new(
-        "monitor-removed",
+    signals[SCREEN_REMOVED] = g_signal_new(
+        "screen-removed",
         G_TYPE_FROM_CLASS(displayClass),
         G_SIGNAL_RUN_LAST,
         0, nullptr, nullptr,
         g_cclosure_marshal_generic,
         G_TYPE_NONE, 1,
-        WPE_TYPE_MONITOR);
+        WPE_TYPE_SCREEN);
 }
 
 WPEView* wpeDisplayCreateView(WPEDisplay* display)
 {
     auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
-    return wpeDisplayClass->create_view ? wpeDisplayClass->create_view(display) : nullptr;
+    return wpeDisplayClass->create_view(display);
 }
 
 bool wpeDisplayCheckEGLExtension(WPEDisplay* display, const char* extensionName)
@@ -143,6 +155,12 @@ bool wpeDisplayCheckEGLExtension(WPEDisplay* display, const char* extensionName)
     return addResult.iterator->value;
 }
 
+WPEInputMethodContext* wpeDisplayCreateInputMethodContext(WPEDisplay* display)
+{
+    auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
+    return wpeDisplayClass->create_input_method_context ? wpeDisplayClass->create_input_method_context(display) : wpeInputMethodContextNoneNew();
+}
+
 /**
  * wpe_display_get_default:
  *
@@ -167,9 +185,10 @@ WPEDisplay* wpe_display_get_default(void)
                     s_defaultDisplay = WTFMove(display);
                     return;
                 }
-                g_warning("Failed to connect to display of type %s: %s", extensionName, error->message);
+                g_error("Failed to connect to display of type %s: %s", extensionName, error->message);
             } else
-                g_warning("Display of type %s was not found", extensionName);
+                g_error("Display of type %s was not found", extensionName);
+            return;
         }
 
         auto* extensionList = g_io_extension_point_get_extensions(extensionPoint);
@@ -185,6 +204,36 @@ WPEDisplay* wpe_display_get_default(void)
     return s_defaultDisplay.get();
 }
 
+/**
+ * wpe_display_get_primary:
+ *
+ * Get the primary display. By default, the first #WPEDisplay that is
+ * created is set as primary display. This is the desired behavior in
+ * most of the cases, but you can set any #WPEDisplay as primary
+ * calling wpe_display_set_primary() if needed.
+ *
+ * Returns: (nullable) (transfer none): the primary display or %NULL
+ */
+WPEDisplay* wpe_display_get_primary()
+{
+    return s_primaryDisplay.get();
+}
+
+/**
+ * wpe_display_set_primary:
+ * @display: a #WPEDisplay
+ *
+ * Set @display as the primary display.
+ *
+ * In most of the cases you don't need to call this, because
+ * the first #WPEDisplay that is created is set as primary
+ * automatically.
+ */
+void wpe_display_set_primary(WPEDisplay* display)
+{
+    s_primaryDisplay.reset(display);
+}
+
 /**
  * wpe_display_connect
  * @display: a #WPEDisplay
@@ -255,6 +304,9 @@ gpointer wpe_display_get_egl_display(WPEDisplay* display, GError** error)
  *
  * Get the #WPEKeymap of @display
  *
+ * As a fallback, a #WPEKeymapXKB for the pc105 "US" layout is returned if the actual display
+ * implementation does not provide a keymap itself.
+ *
  * Returns: (transfer none): a #WPEKeymap or %NULL in case of error
  */
 WPEKeymap* wpe_display_get_keymap(WPEDisplay* display, GError** error)
@@ -263,15 +315,17 @@ WPEKeymap* wpe_display_get_keymap(WPEDisplay* display, GError** error)
 
     auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
     if (!wpeDisplayClass->get_keymap) {
-        g_set_error_literal(error, WPE_DISPLAY_ERROR, WPE_DISPLAY_ERROR_NOT_SUPPORTED, "Operation not supported");
-        return nullptr;
+        auto* priv = display->priv;
+        if (!priv->keymap)
+            priv->keymap = adoptGRef(wpe_keymap_xkb_new());
+        return priv->keymap.get();
     }
 
     return wpeDisplayClass->get_keymap(display, error);
 }
 
 #if USE(LIBDRM)
-static GList* wpeDisplayPreferredDMABufFormats(WPEDisplay* display)
+static GRefPtr<WPEBufferDMABufFormats> wpeDisplayPreferredDMABufFormats(WPEDisplay* display)
 {
     auto eglDisplay = static_cast<EGLDisplay>(wpe_display_get_egl_display(display, nullptr));
     if (!eglDisplay)
@@ -296,27 +350,29 @@ static GList* wpeDisplayPreferredDMABufFormats(WPEDisplay* display)
     static PFNEGLQUERYDMABUFMODIFIERSEXTPROC s_eglQueryDmaBufModifiersEXT = wpeDisplayCheckEGLExtension(display, "EXT_image_dma_buf_import_modifiers") ?
         reinterpret_cast<PFNEGLQUERYDMABUFMODIFIERSEXTPROC>(eglGetProcAddress("eglQueryDmaBufModifiersEXT")) : nullptr;
 
-    GList* preferredFormats = nullptr;
+    auto modifiersForFormat = [&](EGLint format) -> Vector<EGLuint64KHR> {
+        if (!s_eglQueryDmaBufModifiersEXT)
+            return { DRM_FORMAT_MOD_INVALID };
+
+        EGLint modifiersCount;
+        if (!s_eglQueryDmaBufModifiersEXT(eglDisplay, format, 0, nullptr, nullptr, &modifiersCount) || !modifiersCount)
+            return { DRM_FORMAT_MOD_INVALID };
+
+        Vector<EGLuint64KHR> modifiers(modifiersCount);
+        if (!s_eglQueryDmaBufModifiersEXT(eglDisplay, format, modifiersCount, reinterpret_cast<EGLuint64KHR*>(modifiers.data()), nullptr, &modifiersCount))
+            return { DRM_FORMAT_MOD_INVALID };
+
+        return modifiers;
+    };
+
+    auto* builder = wpe_buffer_dma_buf_formats_builder_new(wpe_display_get_drm_render_node(display));
+    wpe_buffer_dma_buf_formats_builder_append_group(builder, nullptr, WPE_BUFFER_DMA_BUF_FORMAT_USAGE_RENDERING);
     for (auto format : formats) {
-        GRefPtr<GArray> dmabufModifiers = adoptGRef(g_array_sized_new(FALSE, TRUE, sizeof(guint64), 1));
-        guint64 invalidModifier = DRM_FORMAT_MOD_INVALID;
-        g_array_append_val(dmabufModifiers.get(), invalidModifier);
-        if (s_eglQueryDmaBufModifiersEXT) {
-            EGLint modifiersCount;
-            if (s_eglQueryDmaBufModifiersEXT(eglDisplay, format, 0, nullptr, nullptr, &modifiersCount) && modifiersCount) {
-                Vector<EGLuint64KHR> modifiers(modifiersCount);
-                if (s_eglQueryDmaBufModifiersEXT(eglDisplay, format, modifiersCount, reinterpret_cast<EGLuint64KHR*>(modifiers.data()), nullptr, &modifiersCount)) {
-                    g_array_set_size(dmabufModifiers.get(), modifiersCount);
-                    for (int i = 0; i < modifiersCount; ++i) {
-                        guint64* modifier = &g_array_index(dmabufModifiers.get(), guint64, i);
-                        *modifier = modifiers[i];
-                    }
-                }
-            }
-        }
-        preferredFormats = g_list_prepend(preferredFormats, wpe_buffer_dma_buf_format_new(WPE_BUFFER_DMA_BUF_FORMAT_USAGE_RENDERING, static_cast<guint32>(format), dmabufModifiers.get()));
+        auto modifiers = modifiersForFormat(format);
+        for (auto modifier : modifiers)
+            wpe_buffer_dma_buf_formats_builder_append_format(builder, format, modifier);
     }
-    return g_list_reverse(preferredFormats);
+    return adoptGRef(wpe_buffer_dma_buf_formats_builder_end(builder));
 }
 #endif
 
@@ -326,9 +382,9 @@ static GList* wpeDisplayPreferredDMABufFormats(WPEDisplay* display)
  *
  * Get the list of preferred DMA-BUF buffer formats for @display.
  *
- * Returns: (transfer none) (element-type WPEBufferDMABufFormat) (nullable): a #GList of #WPEBufferDMABufFormat
+ * Returns: (transfer none) (nullable): a #WPEBufferDMABufFormats
  */
-GList* wpe_display_get_preferred_dma_buf_formats(WPEDisplay* display)
+WPEBufferDMABufFormats* wpe_display_get_preferred_dma_buf_formats(WPEDisplay* display)
 {
     g_return_val_if_fail(WPE_IS_DISPLAY(display), nullptr);
 
@@ -336,7 +392,7 @@ GList* wpe_display_get_preferred_dma_buf_formats(WPEDisplay* display)
     if (!priv->preferredDMABufFormats) {
         auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
         if (wpeDisplayClass->get_preferred_dma_buf_formats)
-            priv->preferredDMABufFormats = wpeDisplayClass->get_preferred_dma_buf_formats(display);
+            priv->preferredDMABufFormats = adoptGRef(wpeDisplayClass->get_preferred_dma_buf_formats(display));
 
 #if USE(LIBDRM)
         if (!priv->preferredDMABufFormats)
@@ -344,78 +400,78 @@ GList* wpe_display_get_preferred_dma_buf_formats(WPEDisplay* display)
 #endif
     }
 
-    return display->priv->preferredDMABufFormats;
+    return display->priv->preferredDMABufFormats.get();
 }
 
 /**
- * wpe_display_get_n_monitors:
+ * wpe_display_get_n_screens:
  * @display: a #WPEDisplay
  *
- * Get the number of monitors of @display
+ * Get the number of screens of @display
  *
- * Returns: the number of monitors
+ * Returns: the number of screens
  */
-guint wpe_display_get_n_monitors(WPEDisplay* display)
+guint wpe_display_get_n_screens(WPEDisplay* display)
 {
     g_return_val_if_fail(WPE_IS_DISPLAY(display), 0);
 
     auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
-    if (!wpeDisplayClass->get_n_monitors)
+    if (!wpeDisplayClass->get_n_screens)
         return 0;
 
-    return wpeDisplayClass->get_n_monitors(display);
+    return wpeDisplayClass->get_n_screens(display);
 }
 
 /**
- * wpe_display_get_monitor:
+ * wpe_display_get_screen:
  * @display: a #WPEDisplay
- * @index: the number of the monitor
+ * @index: the number of the screen
  *
- * Get the monitor of @display at @index
+ * Get the screen of @display at @index
  *
- * Returns: (transfer none) (nullable): a #WPEMonitor, or %NULL
+ * Returns: (transfer none) (nullable): a #WPEScreen, or %NULL
  */
-WPEMonitor* wpe_display_get_monitor(WPEDisplay* display, guint index)
+WPEScreen* wpe_display_get_screen(WPEDisplay* display, guint index)
 {
     g_return_val_if_fail(WPE_IS_DISPLAY(display), nullptr);
 
     auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
-    if (!wpeDisplayClass->get_monitor)
+    if (!wpeDisplayClass->get_screen)
         return nullptr;
 
-    return wpeDisplayClass->get_monitor(display, index);
+    return wpeDisplayClass->get_screen(display, index);
 }
 
 /**
- * wpe_display_monitor_added:
+ * wpe_display_screen_added:
  * @display: a #WPEDisplay
- * @monitor: the #WPEMonitor added
+ * @screen: the #WPEScreen added
  *
- * Emit the signal #WPEDisplay::monitor-added.
+ * Emit the signal #WPEDisplay::screen-added.
  */
-void wpe_display_monitor_added(WPEDisplay* display, WPEMonitor* monitor)
+void wpe_display_screen_added(WPEDisplay* display, WPEScreen* screen)
 {
     g_return_if_fail(WPE_IS_DISPLAY(display));
-    g_return_if_fail(WPE_IS_MONITOR(monitor));
+    g_return_if_fail(WPE_IS_SCREEN(screen));
 
-    g_signal_emit(display, signals[MONITOR_ADDED], 0, monitor);
+    g_signal_emit(display, signals[SCREEN_ADDED], 0, screen);
 }
 
 /**
- * wpe_display_monitor_removed:
+ * wpe_display_screen_removed:
  * @display: a #WPEDisplay
- * @monitor: the #WPEMonitor removed
+ * @screen: the #WPEScreen removed
  *
- * Emit the signal #WPEDisplay::monitor-removed.
- * Note that wpe_monitor_invalidate() is called before the signal is emitted.
+ * Emit the signal #WPEDisplay::screen-removed.
+ * Note that wpe_screen_invalidate() is called before the signal is emitted.
  */
-void wpe_display_monitor_removed(WPEDisplay* display, WPEMonitor* monitor)
+void wpe_display_screen_removed(WPEDisplay* display, WPEScreen* screen)
 {
     g_return_if_fail(WPE_IS_DISPLAY(display));
-    g_return_if_fail(WPE_IS_MONITOR(monitor));
+    g_return_if_fail(WPE_IS_SCREEN(screen));
 
-    wpe_monitor_invalidate(monitor);
-    g_signal_emit(display, signals[MONITOR_REMOVED], 0, monitor);
+    wpe_screen_invalidate(screen);
+    g_signal_emit(display, signals[SCREEN_REMOVED], 0, screen);
 }
 
 static bool isSotfwareRast()
@@ -436,106 +492,69 @@ static bool isSotfwareRast()
     return swrast;
 }
 
-static CString& renderNodeFile()
+/**
+ * wpe_display_get_drm_device:
+ * @display: a #WPEDisplay
+ *
+ * Get the DRM device of @display.
+ *
+ * Returns: (transfer none) (nullable): the filename of the DRM device node, or %NULL
+ */
+const char* wpe_display_get_drm_device(WPEDisplay* display)
 {
-    static NeverDestroyed<CString> renderNode;
-    return renderNode.get();
+    g_return_val_if_fail(WPE_IS_DISPLAY(display), nullptr);
+
+    if (isSotfwareRast())
+        return nullptr;
+
+    static const char* envDeviceFile = getenv("WPE_DRM_DEVICE");
+    if (envDeviceFile && *envDeviceFile)
+        return envDeviceFile;
+
+    auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
+    return wpeDisplayClass->get_drm_device ? wpeDisplayClass->get_drm_device(display) : nullptr;
 }
 
-#if USE(LIBDRM)
-static const CString wpeDRMRenderNodeFile()
-{
-    drmDevicePtr devices[64];
-    memset(devices, 0, sizeof(devices));
-
-    int numDevices = drmGetDevices2(0, devices, std::size(devices));
-    if (numDevices <= 0)
-        return { };
-
-    CString renderNodeDeviceFile;
-    for (int i = 0; i < numDevices && renderNodeDeviceFile.isNull(); ++i) {
-        drmDevice* device = devices[i];
-        if (!(device->available_nodes & (1 << DRM_NODE_RENDER)))
-            continue;
-
-        renderNodeDeviceFile = device->nodes[DRM_NODE_RENDER];
-    }
-    drmFreeDevices(devices, numDevices);
-
-    return renderNodeDeviceFile;
-}
-#endif
-
 /**
- * wpe_render_node_device:
+ * wpe_display_get_drm_render_node:
+ * @display: a #WPEDisplay
  *
- * Get the render node device to be used for rendering.
+ * Get the DRM render node of @display.
  *
- * Returns: (transfer none) (nullable): a render node device, or %NULL.
+ * Returns: (transfer none) (nullable): the filename of the DRM render node, or %NULL
  */
-const char* wpe_render_node_device()
+const char* wpe_display_get_drm_render_node(WPEDisplay* display)
 {
-    auto& renderNode = renderNodeFile();
-    if (renderNode.isNull() && !isSotfwareRast()) {
-        const char* envDeviceFile = getenv("WPE_RENDER_NODE_DEVICE");
-        if (envDeviceFile && *envDeviceFile)
-            renderNode = envDeviceFile;
-#if USE(LIBDRM)
-        else
-            renderNode = wpeDRMRenderNodeFile();
-#endif
-    }
-    return renderNode.data();
+    g_return_val_if_fail(WPE_IS_DISPLAY(display), nullptr);
+
+    if (isSotfwareRast())
+        return nullptr;
+
+    static const char* envDeviceFile = getenv("WPE_DRM_RENDER_NODE");
+    if (envDeviceFile && *envDeviceFile)
+        return envDeviceFile;
+
+    auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
+    return wpeDisplayClass->get_drm_render_node ? wpeDisplayClass->get_drm_render_node(display) : nullptr;
 }
 
-static CString& renderDeviceFile()
-{
-    static NeverDestroyed<CString> renderDevice;
-    return renderDevice.get();
-}
-
-#if USE(LIBDRM)
-static const CString wpeDRMRenderDeviceFile()
-{
-    drmDevicePtr devices[64];
-    memset(devices, 0, sizeof(devices));
-
-    int numDevices = drmGetDevices2(0, devices, std::size(devices));
-    if (numDevices <= 0)
-        return { };
-
-    CString renderDeviceFile;
-    for (int i = 0; i < numDevices && renderDeviceFile.isNull(); ++i) {
-        drmDevice* device = devices[i];
-        if (!(device->available_nodes & (1 << DRM_NODE_PRIMARY | 1 << DRM_NODE_RENDER)))
-            continue;
-
-        renderDeviceFile = device->nodes[DRM_NODE_PRIMARY];
-    }
-    drmFreeDevices(devices, numDevices);
-
-    return renderDeviceFile;
-}
-#endif
-
 /**
- * wpe_render_device:
+ * wpe_display_use_explicit_sync:
+ * @display: a #WPEDisplay
  *
- * Get the render device to be used for rendering.
+ * Get whether explicit sync should be used with @display for
+ * supported buffers.
  *
- * Returns: (transfer none) (nullable): a render device, or %NULL.
+ * Returns: %TRUE if explicit sync should be used, or %FALSE otherwise
  */
-const char* wpe_render_device()
+gboolean wpe_display_use_explicit_sync(WPEDisplay* display)
 {
-    auto& renderDevice = renderDeviceFile();
-    if (renderDevice.isNull() && !isSotfwareRast()) {
-        const char* envDeviceFile = getenv("WPE_RENDER_DEVICE");
-        if (envDeviceFile && *envDeviceFile)
-            renderDevice = envDeviceFile;
-#if USE(LIBDRM)
-        else
-            renderDevice = wpeDRMRenderDeviceFile();
-#endif
-    }
-    return renderDevice.data();
+    g_return_val_if_fail(WPE_IS_DISPLAY(display), FALSE);
+
+    static const char* envExplicitSync = getenv("WPE_USE_EXPLICIT_SYNC");
+    if (envExplicitSync && !strcmp(envExplicitSync, "0"))
+        return false;
+
+    auto* wpeDisplayClass = WPE_DISPLAY_GET_CLASS(display);
+    return wpeDisplayClass->use_explicit_sync ? wpeDisplayClass->use_explicit_sync(display) : FALSE;
 }
